
import copy

from . import devices

# PFM_START = 0x0000
# USER_ID_START = 0x8000
# CONFIG_WORD_START = 0x8007


class HexfileDecoder:

    def __init__(self, path, device):
        self.path = path

        self.device = devices.Device(device)
        self.device.readDeviceFile(devices.XC8Manager.findXC8Installs()[0])

        self.ascii_records = self._readHexfile(path)
        self.records = self._decodeAsciiRecords(self.ascii_records)
        self.word_list = self._decodeWordsFromRecords(self.records)

        self.userflash_words = [addr for addr in self.word_list.keys() if addr <= self.device.flash.end]
        self.config_words = [addr for addr in self.word_list.keys() if addr > self.device.flash.end]
        
        self.memory = self._separateRows(self.word_list)

    
    def getMemory(self):
        return self.memory

    def printMemory(self):
        print('\n PFM   :   x0     x1     x2     x3     x4     x5     x6     x7     x8     x9     xA     xB     xC     xD     xE     xF')
        print('-------:-----------------------------------------------------------------------------------------------------------------')
        for word_address in self.userflash_words:
            if word_address % 16 == 0 or word_address == 0:
                if word_address % 64 == 0:
                    if word_address == 0:
                        print('0x%.4X + ' % word_address, end='')
                    else:
                        print('\n0x%.4X + ' % word_address, end='')
                else:
                    print('\n0x%.4X | ' % word_address, end='')
            print('0x%.4X ' % self.word_list[word_address], end='')

        print('\n       :\nCONFIG :\n-------:--------')
        for word_address in self.config_words:
            print ('0x%.4X | 0x%.4X' % (word_address, self.word_list[word_address]))
        print('')

    # Read hexfile in and output a list of records
    # A record is an intel hex 'command'
    # Each record is proceeded by an ascii ':'
    def _readHexfile(self, path_to_hexfile):
        # Read the entire hexfile (Shouldn't be more than 56k * 2)
        with open(path_to_hexfile) as hexfile:
            hexfile_data = hexfile.read()

        hexfile_data = hexfile_data.replace('\n', '') # Remove newlines
        hexfile_data = hexfile_data.lstrip(':').split(':') # Split the file at the record marks ':' to get a list of records
        return hexfile_data

    # Decode the list of ascii records to a list of dicts containing record information
    def _decodeAsciiRecords(self, ascii_records):
        decoded_records = []
        for record in ascii_records:
            data_len = int(record[0:2], base=16) # First ASCII hex byte is the data length
            offset_addr = int(record[2:6], base=16) # Next two ASCII hex bytes is the offset address
            record_type = int(record[6:8], base=16) # Next byte is the record type
            data = record[8:(data_len*2)+8] # The data is data_len*2 long since 2 ascii chars represent one hex byte
            checksum = int(record[-2:], base=16) # The Last byte in the record is the checksum
            decoded_records.append(dict(data_len=data_len, offset_addr=offset_addr, record_type=record_type, data=data, checksum=checksum))
        return decoded_records

    # Decodes a list of record objects to a dictionary containg [Address]: <Word>
    # Records MUST be in the order they were in the hexfile
    # Hex records supported so far are:
    # - DATA: 0x00
    # - Ext Linear Address: 0x04
    def _decodeWordsFromRecords(self, decoded_records):
        words = {}
        high_address = 0 # The high address defaults to 0x0000 unless a hex record sets it otherwise
        for record in decoded_records: 
            if record['offset_addr'] != 0:
                low_address = record['offset_addr'] // 2    # Address are 2x that of the pic memory in the hex file since it takes 2 bytes per word
            else:
                low_address = 0
            word_start = 0

            if record['record_type'] == 4 and record['data_len'] == 2: # A record with type 4 sets the high address
                high_address = int(record['data'], base=16)
                high_address = (high_address << 16) // 2    # Address are double in the hex file

            elif record['record_type'] == 0: # A record with type 0 is a data record that holds words
                # Loop through our 'data' and extract the words while calculating a direct address
                while word_start <= (record['data_len'] * 2) - 4:
                    # The complete address is a combination of 2 high bytes and 2 low bytes which represent a 15-bit address
                    address = high_address + low_address
                    # The ascii hex representation of a word is MSB second, LSB byte first. So we swap those around and convert it to a number
                    word = int(record['data'][word_start+2:word_start+4] + record['data'][word_start:word_start+2], base=16)
                    words[address] = word

                    # Skip to the next word and increment the address
                    word_start += 4
                    low_address += 1

        return words

    # Separates a dict of {addr: word} into rows 64 words long
    # Returns a dict of {start_addr: [row_data]} with row data
    # being a list of words paded with 0x3FFF
    def _separateRows(self, words):
        rows = {}

        for word_address in self.userflash_words:
            
            row_start_address = word_address - (word_address % 64)
            row_address_offset = word_address - row_start_address
            
            if row_start_address not in rows:
                rows[row_start_address] = [0x3FFF for _ in range(64)]
            
            rows[row_start_address][row_address_offset] = words[word_address]
        
        for word_address in self.config_words:
            rows[word_address] = [words[word_address]]
        
        return rows




import argparse
import os.path
import sys

from . import hexfile
from . import programmer


DESCRIPTION = '''\
A utility to aid in programming PIC microcontrollers\
'''

USAGE = '''\
picchick [options] [hexfile]\
'''

EPILOG = '''\
flag arguments:
  [addr]:\t\tdevice memory address in hexadecimal
\tall\t\tall device memory areas
\tflash\t\tuser flash area
'''

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=DESCRIPTION,
    usage=USAGE,
    epilog=EPILOG)

parser.add_argument('hexfile',
    nargs='?',
    default=None,
    help='path to the hexfile')

parser.add_argument('-f', '--flash',
    action='store_true',
    help='flash hexfile onto the device')
parser.add_argument('--read',
    metavar='addr',
    help='read specified address or chunk  of memory')
parser.add_argument('--write',
    nargs=2,
    metavar=('addr', 'word'),
    help='write word to specified address')
parser.add_argument('--erase',
    nargs='?',
    const='all',
    metavar='addr',
    help='erase device or specified address')

parser.add_argument('-d', '--device',
    metavar='chipID',
    help='device to be programmed')
parser.add_argument('-p', '--port',
    metavar='port',
    help='programmer serial port')
parser.add_argument('--baud',
    type=int,
    default=9600,
    metavar='baud',
    help='serial connection baudrate',)

parser.add_argument('--map',
    action='store_true',
    help='display the hexfile')
parser.add_argument('--list-ports',
    action='store_true',
    help='list available serial ports')
parser.add_argument('--list-devices',
    action='store_true',
    help='list available device configurations')

def parseArgv():
    args = parser.parse_args()

    # Requirements tree

    # Flash flag requires both the hexfile and the programmer (port)
    both_reqd = (args.flash)
    # The read and erase flags only require the programmer connection (and port flag)
    programmer_reqd = both_reqd or (args.read or args.erase or args.write)
    # The map flag only requires the hexfile to be present
    hexfile_reqd = both_reqd or (args.map)

    # If we don't need to do anything, print help because
    # the user needs it
    if not hexfile_reqd and not programmer_reqd:
        parser.print_help()
        sys.exit(0)


    # Firstly, if we need the hexfile, check if it exists
    # If not, immediatly exit with a helpful message
    if hexfile_reqd:
        if args.hexfile is None:
            print(f"Missing argument: hexfile")
            sys.exit(1)
        elif args.device is None:
            print("Missing argument: -d, --device chipID")
        elif not os.path.isfile(args.hexfile):
            print(f"Could not find hexfile: { args.hexfile}")
            sys.exit(1)
        else:
            print(f"Using hexfile: { args.hexfile }")
        hex_decoder = hexfile.HexfileDecoder(args.hexfile, args.device)

    # We now have all the hexfile reqs, so take care of the actions
    # that only require the hexfile
    if args.map:
        hex_decoder.printMemory()


    # Second if we need the programmer, we check:
    # - If the -p argument is specified
    # - If the path exists (This may not work on windows)
    if programmer_reqd:
        if args.port is None:
            print("Missing argument: -p port")
            sys.exit(1)
        elif not os.path.exists(args.port):
            print(f"Could not find port: { args.port }")
            sys.exit(1)
        else:
            dev = programmer.Programmer(args.port, baud=args.baud)
            if not dev.connect():
                print(f"ERROR: Failed to connect to device: { args.port } Exiting...")
                sys.exit(1)


    # We now have all the programmer reqs, so do the actions that only
    # need the programmer:
    # Display information about ports if flag was included
    if args.list_ports:
        # if args.port is None:
        programmer.listPorts()
        # else:
            # print("Not including port-list due to valid port flag")


    if args.erase or args.flash or args.read or args.write:
        dev.start()

        if args.erase:
            if args.erase == 'all':
                dev.erase(0xFFFF)
            elif args.erase == 'flash':
                dev.erase(0xEFFF)
            else:
                dev.erase(int(args.erase, base=16))
        
        if args.flash:
            

            for start_address, row in hex_decoder.memory.items():
                if start_address < hexfile.USER_ID_START:
                    dev.row(start_address, row)
            for start_address, word in hex_decoder.memory.items():
                if hexfile.USER_ID_START <= start_address < hexfile.CONFIG_WORD_START:
                    dev.word(start_address, word[0])
            for start_address, word in hex_decoder.memory.items():
                if hexfile.CONFIG_WORD_START <= start_address:
                    dev.word(start_address, word[0])
        elif args.write:
            dev.word(int(args.write[0], base=16), int(args.write[1], base=16))
        
        if args.read:
            dev.read(int(args.read, base=16))
        dev.stop() 
        dev.disconnect()



import configparser
import pathlib


class MemoryRange:

    size = None
    start = None
    end = None

    def configure(self, size=None, addr_range=None):

        if size and not addr_range:
            self.size = int(size, base=16)
            self.start = 0
            self.end = self.size - 1

        if addr_range and not size:
            self.start = int(addr_range.split('-')[0], base=16)
            self.end = int(addr_range.split('-')[1], base=16)
            self.size = self.end - self.start + 1


# This class holds the device data needed for flashing etc.
class Device:

    flash = MemoryRange()
    config = MemoryRange()

    def __init__(self, chip_id):

        self.chip_id = chip_id.upper()
    
    def readDeviceFile(self, ccpath):
        self.device_file_path = ccpath / 'pic/dat/ini' / (self.chip_id.lower() + '.ini')
        self.device_file = configparser.ConfigParser(strict=False)
        self.device_file.read(self.device_file_path)
        self.flash.configure(size=self.device_file.get(self.chip_id, 'ROMSIZE'))
        self.config.configure(addr_range=self.device_file.get(self.chip_id, 'CONFIG'))


XC8_COMMON_PATHS = [
    '/opt/microchip/xc8'
]

# This class handles searching the local filesystem for the xc8 compiler so
# we can use its device files
class XC8Manager:

    # xc8_paths = []
    
    def findXC8Installs():

        found_installs = []

        for common_path in XC8_COMMON_PATHS:
            path = pathlib.Path(common_path)

            if path.is_dir():
                found_installs += [p for p in path.iterdir() if p.is_dir()]
        
        # self.xc8_paths += found_installs
        return found_installs








import serial
import serial.tools.list_ports


# Helper functions to deal with ascii and binary
def ASCII(string):
    return string.encode(encoding='ascii')

def INTBYTES(number, len=2):
    return number.to_bytes(len, 'big')

def ROWBYTES(row):
    word_bytes = bytearray()
    for word in row:
        word_bytes += INTBYTES(word)
    return word_bytes


# PreDefined Characters and Commands
SEP = ASCII(':')
OK = ASCII('OK')
GREETING = ASCII('HELLO')
BYE = ASCII('BYE')
START = ASCII('START')
STOP = ASCII('STOP')
ADDR = ASCII('ADDR')
ROW = ASCII('ROW')
WORD = ASCII('WORD')
READ = ASCII('READ')
ERASE = ASCII('ERASE')


# General Helper functions
def wait_print(string):
    print(string, end=' ', flush=True)



class Programmer:

    def __init__(self, port, baud=9600, timeout=30):
        self._conn = serial.Serial(timeout=timeout)
        self._port = self._conn.port = port
        self._baud = self._conn.baud = baud
    
    def connect(self):
        try:
            self._conn.open()
        except serial.SerialException:
            print(f"Failed to open serial port: { self._port }")
            return False
        wait_print(f"Connecting to device: { self._port }\nSending greeting...")
        self._conn.flushInput()
        self._conn.write(GREETING + SEP)
        if self.__check_response(expected_resp=GREETING) is not True:
            print('device failed to respond')
            self.disconnect()
            return False
        print("connected to programmer")
        return True
    
    def disconnect(self):
        wait_print('Disconnecting from programmer...')
        self._conn.write(BYE + SEP)
        resp = self.__check_response(expected_resp=BYE)
        self._conn.close()
        if resp is not True:
            print('GOODBYE')
            return False
        print('goodbye')
        return True

    def start(self):
        wait_print('Entering programming mode...')
        self._conn.flushInput()
        self._conn.write(START + SEP)
        if self.__check_response() is not True:
            print('failed. Closing connection')
            self.disconnect()
            return False
        print('success')
        return True
    
    def stop(self):
        wait_print('Leaving programming mode...')
        # print(self._conn.read_all())
        self._conn.flushInput()
        self._conn.write(STOP + SEP)
        if self.__check_response() is not True:
            print('failed. Closing connection')
            self.disconnect()
            return False
        print ('success')
        return True
    
    def address(self, address):
        wait_print("Setting start address to: 0x%.4X..." % address)
        self._conn.write(ADDR + SEP + INTBYTES(address))
        if self.__check_response() is not True:
            print('failed. Closing connection')
            self.disconnect()
            return False
        print('success')
        return True
    
    def word(self, address, word):
        wait_print("Writing Word: 0x%.4X | 0x%.4X..." % (address, word))
        self._conn.write(WORD + SEP + INTBYTES(address) + SEP + INTBYTES(word))
        if self.__check_response() is not True:
            print('failed')
            return False
        print('success')
        return True
    
    def row(self, address, row):
        wait_print("Writing Row: 0x%.4X..." % (address))
        cmd = ROW + SEP + INTBYTES(address) + SEP + ROWBYTES(row)
        # print(cmd)
        self._conn.write(cmd)
        if self.__check_response() is not True:
            print('failed')
            return False
        print('success')
        return True

    def read(self, address):
        wait_print("Reading from address: 0x%.4X..." % (address))
        self._conn.write(READ + SEP + INTBYTES(address))
        if self.__check_response() is not True:
            print('failed')
            return False
        resp = self._conn.read(size=2)
        print('0x' + resp.hex().upper())
        return True
    
    def erase(self, address):
        wait_print("Erasing Row: 0x%.4X..." % (address))
        self._conn.write(ERASE + SEP + INTBYTES(address))
        if self.__check_response() is not True:
            print('failed')
            return False
        print('success')
        return True


    def __check_response(self, expected_resp=OK):
        resp = self._conn.read_until(expected=SEP)
        # print(resp)
        if resp.rstrip(SEP) == expected_resp:
            return True
        else:
            return resp


# Utlity functions
def listPorts():
    ports = serial.tools.list_ports.comports()
    print(f"{ len(ports) } serial devices found:")
    for port in ports:
        print(f"{ port.device }\t{ port.product }")
